from .timer import TimerManager
from .service import ServiceManager

from dataclasses import dataclass


@dataclass(kw_only=True)
class ServiceTimerManager:
    # todo: manage directory when you put services and timers.
    service_filename: str 
    service_description: str = ""
    command: str

    timer_filename: str
    timer_description: str = ""
    on_calendar: str
    
    overwrite: bool = False

    def schedule(self):
        self._create_service()
        self._create_timer()
        self.timer_manager.start_timer()

    def _create_service(self):
        self.service_manager = ServiceManager(
            filename=self.service_filename, 
            description=self.service_description, 
            command=self.command,
            overwrite=self.overwrite,
        )
        self.service_manager.create_service_file()

    def _create_timer(self):
        self.timer_manager = TimerManager(
            filename=self.timer_filename,
            description=self.timer_description,
            on_calendar=self.on_calendar,
            service_manager=self.service_manager,
            overwrite=self.overwrite,
        )
        self.timer_manager.create_timer()


from .file_manager import FileManager
from . import config


class ServiceManager:
    """Service class manages: create the service file
    Args:
        filename (string): name of the service file with extension
        command (string): command to run
    """

    def __init__(
        self,
        filename: str,
        command: str,
        description: str = "",
        overwrite: bool = False,
    ):
        self.filename: str = filename
        # todo: set a default description: title or the first part of filename
        # todo: "title" service
        self.description = description
        self.command = command
        self.overwrite = overwrite

        self.logger = config.get_logger(__name__)
        self.logger.info("create an instance of ServiceManager")

    def create_service_file(self):
        self.logger.info("create service file")
        self.file_manager = FileManager(self.filename, overwrite=self.overwrite)
        self.file_manager.create_file(content=self._get_service_text())
        self.logger.info("created service file")

    def _get_service_text(self):
        return f"""[Unit]
Description={self.description}


[Service]
User=root
ExecStart={self.command}
"""


if __name__ == "__main__":
    service_manager = ServiceManager(
        filename="test.service", command="screensaver-command -l"
    )
    service_manager.create_service_file()



from .service_timer_manager import ServiceTimerManager
from . import config

class SimpleScheduler:
    def __init__(self, title: str, command: str, on_calendar: str, overwrite: bool = False):
        self.logger = config.get_logger(__class__.__name__)
        self.logger.info("create instance of SimpleScheduler")

        self.title = title
        self.command = command
        self.on_calendar = on_calendar
        self.overwrite = overwrite
    
    def schedule(self):
        self.logger.info("create instance of ServiceTimerManager")
        self.service_timer_manager = ServiceTimerManager(
            service_filename=f"{self.title.replace(' ', '_')}.service",
            service_description=self.title,
            command=self.command,
            timer_filename=f"{self.title.replace(' ', '_')}.timer",
            timer_description=self.title,
            on_calendar=self.on_calendar,
            overwrite=self.overwrite,
        )
        self.logger.info("created instance of ServiceTimerManager")
        self.service_timer_manager.schedule()
        self.logger.info("schedule complete")

        
if __name__ == '__main__':
    import os 
    title='testing simple scheduler'
    title = title.replace(' ', '_')
    sc = SimpleScheduler(
        title=title,
        command='shutdown now',
        on_calendar='daily',
        overwrite=True,
    )
    sc.schedule()
    os.remove(f'/etc/systemd/system/{title}.service')
    os.remove(f'/etc/systemd/system/{title}.timer')   


from schedulerx import ServiceTimerManager

service_timer = ServiceTimerManager(
    service_filename="shutdown.service",
    service_description="shutdown at midnight",
    command="shutdown now",
    timer_filename="shutdown.timer",
    timer_description="shutdown at midnight timer",
    on_calendar="@daily",
)


from .permission_manager import PermissionManager

import os
from . import config 


SYSTEMD_SYSTEM_DIR = r"/etc/systemd/system"

ADD_OTHERS_WRITE_PERMISSIONS = "o+w"
REMOVE_OTHERS_WRITE_PERMISSIONS = "o-w"
class FileManager:
    """Class contains methods to manage files in the /etc/systemd/system directory"""


    def __init__(self, filename: str, overwrite=False):
        self.logger = config.get_logger(__name__)
        self.logger.info("create instance of FileManager")

        self.filename = filename
        self.overwrite = overwrite
        self.permission_manager: PermissionManager = PermissionManager()

    @property
    def file_full_path(self):
        return f"{SYSTEMD_SYSTEM_DIR}/{self.filename}"

    def _save_origin_systemd_writable_permission(self):
        """save the origin directory permissions in origin_directory_permission"""
        if os.access(SYSTEMD_SYSTEM_DIR, os.W_OK):
            self.origin_systemd_writable_permission = ADD_OTHERS_WRITE_PERMISSIONS
        else:
            self.origin_systemd_writable_permission = (
                REMOVE_OTHERS_WRITE_PERMISSIONS
            )

    def create_file(self, content):
        """Create a file at /etc/systemd/system/ and add write permission to it

        Args:
            filename (string): filename with extension
            content (string): what to write at /etc/systemd/system/$filename
        """

        self.check_permissions()

        self._save_origin_systemd_writable_permission()

        # add write permission to /etc/systemd/system
        self.logger.debug(
            f"change permission to writable of {SYSTEMD_SYSTEM_DIR} "
        )
        self.permission_manager.change_path_permissions(
            SYSTEMD_SYSTEM_DIR, ADD_OTHERS_WRITE_PERMISSIONS
        )
        self.logger.debug(
            f"changed permission to writable of {SYSTEMD_SYSTEM_DIR} "
        )

        # create the file
        with open(f"{self.file_full_path}", "w") as service_file:
            service_file.write(content)
            self.logger.debug("create and fill file")
            # change permission of file to writable
            self.permission_manager.change_path_permissions(self.file_full_path, "o+w")

        # change /etc/systemd/system permission to it's origin
        self.logger.debug(f"change permission back of {SYSTEMD_SYSTEM_DIR}")
        self.permission_manager.change_path_permissions(
            SYSTEMD_SYSTEM_DIR, self.origin_systemd_writable_permission
        )
        self.logger.info(f"created {self.file_full_path} successfully")

    def is_file_exist(self):
        if os.path.exists(f"{SYSTEMD_SYSTEM_DIR}/{self.filename}"):
            return True
        else:
            return False

    def check_permissions(self):
        if self.is_file_exist():
            if not self.permission_manager.is_writable(self.file_full_path):
                self.logger.error(
                    f"{self.filename} is a system file. try another service name or title"
                )
                raise PermissionError(
                    f"{self.filename} is a system file. try another service name or title"
                )
            else:
                if not self.overwrite:
                    self.logger.error(
                        f"you can't overwrite {self.filename}. try another filename, or set overwrite to True"
                    )
                    raise FileExistsError(
                        f"You can't overwrite {self.filename}. Already existed. If you want to, you can set overwrite to True."
                    )


if __name__ == "__main__":
    file_manager = FileManager("ztest.txt")
    file_manager.create_file("hello world")



from .password_helper import get_root_password
from dataclasses import dataclass
from . import config 

import subprocess

@dataclass
class CommandHandler:
    logger = config.get_logger(__name__)
    logger.info("create instance of CommandHandler")

    def run_shell_command_with_input(self, command: str, input: str=''):
        """run shell command as root .eg. with password as input

        Args:
            command (str): the shell command to run
            input (str): input
        """
        self.logger.debug("run_shell_command_with_input() func is called")
        process = subprocess.run(
            [f"su -c '{command}'"],
            shell=True,
            input=f"{input}\n",
            capture_output=True,
            encoding="utf-8",
            timeout=6,
        )
        self.logger.debug(
            f"process is created: stdout: {process.stdout} || stderr: {process.stderr}"
        )
        return process

    def run_shell_command(self, command: str):
        """ run shell command simple usage"""
        self.logger.debug("run_shell_command() func is called")
        process = subprocess.run(
            [f"{command}"],
            shell=True,
            capture_output=True,
            encoding="utf-8",
            timeout=6,
        )
        self.logger.debug(
            f"process is created: stdout: {process.stdout} || stderr: {process.stderr}"
        )
        return process


    def run_shell_command_as_root(self, command: str):
        """Run a shell command as the root user in Linux.

        Args:
            command (str): The shell command to run.

        Returns:
            None
        """
        self.logger.debug("run_shell_command_as_root() function called")
        if get_root_password() == None:
            raise RuntimeError("ROOT_PASSWORD is not is PATH.")

        try:
            self.run_shell_command_with_input(command, get_root_password())
        except Exception as e: 
            raise RuntimeError(f"Error running shell command: {e}")
        else:
            self.logger.debug(f"Shell command run as root successfully. command: {command}")


if __name__ == "__main__":
    ch = CommandHandler()
    ch.run_shell_command_as_root("notify-send 'hello'")


from .run_command import CommandHandler
from dataclasses import dataclass
from . import config

import os
import stat


@dataclass
class PermissionManager:
    """class that manages linux file permissions"""

    logger = config.get_logger(__name__)
    logger.info("create an instance of PermissionManager")
    command_handler = CommandHandler()

    def change_path_permissions(self, path: str, permissions: str):
        """change a file permission to the desired permission

        Args:
            permissions (str): permission desired like o+w or u-x
            path (str): path to change
        """
        self.logger.debug(
            "change_path_permissions method from PermissionManager is called"
        )
        self.logger.debug(
            f"permissions before changing: {stat.filemode(os.stat(path).st_mode)}"
        )
        self.command_handler.run_shell_command_as_root(
            command=f"chmod {permissions} {path}"
        )
        self.logger.debug(
            f"permissions after changing: {stat.filemode(os.stat(path).st_mode)}"
        )
        self.logger.debug(f"changed permission of {path}")

    def is_writable(self, path: str):
        return True if os.access(path, os.W_OK) else False


if __name__ == "__main__":
    path = "testing/permission_manager.txt"
    with open(path, "w") as f:
        f.write("hello")
    permission_manager = PermissionManager()
    permission_manager.change_path_permissions(path, "o+w")
    os.remove(path)



from .simple_scheduler import SimpleScheduler
from .service_timer_manager import ServiceTimerManager
from .timer import TimerManager
from .service import ServiceManager
from .config import get_logger 
from . import password_helper

__all__ = ["service", "timer", "service_timer_manager", "simple_scheduler"]


import os


def get_root_password():
    """Get the value of the ROOT_PASSWORD environment variable.
    
    Returns:
        str: The value of the ROOT_PASSWORD environment variable.
    """
    return os.environ.get("ROOT_PASSWORD")



from .file_manager import FileManager
from .service import ServiceManager
from .run_command import CommandHandler
from . import config

import os


class TimerManager:
    """

    Args:
        service_manager (ServiceManager): service manager instance linked to
        to this timer instance
        description (str): timer_description
    """

    def __init__(
        self,
        filename: str,
        on_calendar: str,
        service_manager: ServiceManager,
        description: str = "",
        overwrite: bool = False,
    ) -> None:
        self.logger = config.get_logger(__name__)
        self.logger.info("created instance of TimerManager")

        self.filename = filename
        self.description = description
        self.on_calendar = on_calendar
        self.service_manager: ServiceManager = service_manager
        self.overwrite = overwrite

        self.file_manager = FileManager(filename=self.filename, overwrite=self.overwrite)
        self.command_handler = CommandHandler()

    def create_timer(self):
        self.logger.debug(f"create timer file {self.filename}")
        self.file_manager.create_file(self._get_timer_text())
        self.logger.info(f"created timer file {self.filename}")

    def _get_timer_text(self):
        return f"""[Unit]
Description={self.description}
Requires={self.service_manager.filename}


[Timer]
OnCalendar={self.on_calendar}
Unit={self.service_manager.filename}

[Install]
WantedBy=multi-user.target
"""

    def start_timer(self):
        # reload daemon process
        self.logger.debug("reload daemon process")
        self.command_handler.run_shell_command_as_root("systemctl daemon-reload")
        # start timer
        self.logger.debug("start timer")
        self.command_handler.run_shell_command_as_root(
            f"systemctl start {self.filename}"
        )

        self.logger.debug("enable timer")
        self.command_handler.run_shell_command_as_root(
            f"systemctl enable {self.filename}"
        )

        self.logger.info("timer is set successfully")


if __name__ == "__main__":
    title = "tes"
    sm = ServiceManager(f"{title}.service", "ls")
    sm.create_service_file()
    tm = TimerManager(f"{title}.timer", on_calendar="12:12:12", service_manager=sm)
    tm.create_timer()
    os.remove(f"/etc/systemd/system/{title}.service")
    os.remove(f"/etc/systemd/system/{title}.timer")



import logging


def get_logger(name):
    logger = logging.getLogger(name)
    logger.setLevel(logging.DEBUG)

    file_handler = logging.FileHandler(filename="scheduler/log/schedule.log")
    file_handler.setLevel(logging.DEBUG)

    console = logging.StreamHandler()
    console.setLevel(logging.DEBUG)

    # formatter = logging.Formatter("%(asctime)s - module:%(module)s -> %(funcName)s -  %(levelname)s - %(message)s")
    formatter = logging.Formatter("%(asctime)-16s   %(name)-20s -> %(funcName)-30s %(levelname)-8s %(message)s", datefmt="%Y-%m-%d %H:%M")
    console.setFormatter(formatter)
    file_handler.setFormatter(formatter)

    logger.addHandler(console)
    logger.addHandler(file_handler)
    return logger


if __name__ == "__main__":
    pass